#include "V_D_UC.h"
#include "srvr.h"

#ifdef USE_agent

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* CALL:
*   daOpen(name,flags,mode)
*   	char *name;
*   	int   flags;
*   	int   mode;
*
* DESCRIPTION:
*
* Distributed-Access file-open routine.  We pass the  name  to  daPrsPath  to
* split it into a server and file name. We try to contact the server. If that
* succeeds, we send a message to the server asking it to open  the  file  for
* us. If it agrees, we note the fact in our table, and return our file handle
* to the caller.
*
* AUTHOR: John Chambers.
*
* CHANGES:
*/
FCT	daOpen(name,flags,mode)
	char *name;
	int   flags;
	int   mode;
{	int   v = -1;
	int   e = ENOENT;
	Srvr* sp;
static Str srvr = {0};
static Str file = {0};
:f	Fenter("daOpen");
:5	V5D "Called for %s flags=0%o mode=0%o.",Dspp(name),flags,mode D;
	if (!daPrsPath(name,&srvr,&file)) {
		e = errno;
:5		V5D "No server name in %s",Dspp(name) D;
		Fail;
	}
	if (!(sp = daSrvrLink(&srvr))) {
		e = errno;
:5		V5D "Can't contact server %s [Err %d=%s=%s]",DspSV(srvr),Errinfo D;
		Fail;
	}
:5	V5D "Srvr %08X is %s",sp,DspSV(sp->name) D;
	v = daSrvrOpen(sp,&file,flags,mode);
	e = errno;
fail:
:f	Fexit;
	errno = e;
	return v;
}
#endif

#ifdef TEST
/*
* Test driver for Open()
*/
main(ac,av)
	int    ac;
	char** av;
{	int    a, f, s=0;
	int    c0, c1;

:1	ac = Vinit(ac,av);

	for (a=1; a<ac; a++) {
:3		V3D "Arg %d: %s",a,Dspp(av[a]) D;
		c0 = av[a][0];
		Switch(c0) {
		case '-':
		case '+':
			c1 = av[a][1];
:3			V3D "Option: %s",Dspp(av[a]) D;
			Switch(c1) {
			case 'd':
			case 'D':
				Vopt(av[a]+2);
				break;
			default:
				V1D "Unknown option %s ignored.",Dspp(av[a]) D;
				break;
			}
			break;
		default:
:3			V3D "  File: %s",Dspp(av[a]) D;
			if ((f = Open(av[a],0,0)) >= 0) {
				V2D "Success: Open(%s,0,0)=%d",Dspp(av[a]),f D;
			} else {
				V1D "Failure: Open(%s,0,0) [Err %d=%s=%s]",Dspp(av[a]),Errinfo D;
				++s;
			}
			break;
		}
	}
	Exit(s);
}
#endif	/* TEST */
